@@ -6,9 +6,7 @@ This package provides a light-weight interface to use R in Python by pipe.  It
 can be used on multiple platforms since it is written in pure python. 
 
 Prerequisites:
-	1. Python 2.4 or later is required since PypeR uses the python standard
-	   module "subprocess", which was first introduced in Python 2.4.
-	2. On Windows, the pywin32 package is required.
+	1. Python 2.3 or later is required.
 
 Usage:
 	The usage of this packages is very simple. Examples are presented in the
@@ -80,22 +78,79 @@ import time
 import re
 import tempfile
 from types import *
-import subprocess
-import shlex
-from pipeio import Popen, PIPE, send_all, recv_some
 
-if sys.version < '2.3':
+__version__ = '1.1.0'
+
+if sys.version < '2.3': # actually python 2.3 is required by tempfile.mkstemp !!!
 	set = frozenset = tuple
+	basestring = str
 elif sys.version < '2.4':
 	from sets import Set as set, ImmutableSet as frozenset
 
+if sys.version < '3.0':
+	_mystr = _mybytes = lambda s:s
+else:
+	from functools import reduce
+	long, basestring, unicode = int, str, str
+	_mybytes = lambda s:bytes(s, 'utf8') #'ascii')
+	_mystr = lambda s:str(s, 'utf8')
+
 try:
 	import numpy
 	has_numpy = True
 except:
 	has_numpy = False
 
-__version__ = '1.0.4'
+_has_subp = False
+if sys.platform == 'cli': # for IronPython
+	from System.Diagnostics import Process
+	PIPE, _STDOUT = None, None
+
+	def Popen(CMD, *a, **b): 
+		'''
+		CMD is a list - a command and its arguments
+		'''
+		p = Process()
+		p.StartInfo.UseShellExecute = False
+		p.StartInfo.RedirectStandardInput = True
+		p.StartInfo.RedirectStandardOutput = True
+		p.StartInfo.RedirectStandardError = True
+		p.StartInfo.FileName = CMD[0]
+		p.StartInfo.Arguments = ' '.join(CMD[1:])
+		p.Start()
+		return(p)
+
+	def sendAll(p, s):
+		# remove ending newline since WriteLine will add newline at the end of s!
+		if s.endswith('\r\n'): s = s[:-2]
+		elif s.endswith('\n'): s = s[:-1]
+		p.StandardInput.WriteLine(_mybytes(s)) 
+
+	def readLine(p, *a, **b):
+		return(_mystr(p.StandardOutput.ReadLine()) + '\n') # add newline since ReadLine removed it.
+
+else: 
+
+	try:
+		import subprocess
+		_has_subp = True
+		Popen, PIPE, _STDOUT = subprocess.Popen, subprocess.PIPE, subprocess.STDOUT
+	except: # Python 2.3 or older
+		PIPE, _STDOUT = None, None
+		def Popen(CMD, *a, **b):
+			class A: None
+			p = A()
+			p.stdin, p.stdout = os.popen4(' '.join(CMD))
+			return(p)
+
+	def sendAll(p, s):
+		p.stdin.write(_mybytes(s))
+		#os.write(p.stdin.fileno(), s)
+		p.stdin.flush()
+
+	def readLine(p, *a, **b):
+		return(_mystr(p.stdout.readline()))
+
 
 def NoneStr(obj): return 'NULL'
 
@@ -136,7 +191,7 @@ def SeqStr(obj, head='c(', tail=')'):
 	return (is_int and 'as.integer(' or '') + head + ','.join(map(Str4R, obj)) + tail + (is_int and ')' or '') 
 
 def DictStr(obj):
-	return 'list(' + ','.join(map(lambda a:'%s=%s' % (Str4R(a[0]), Str4R(a[1])), obj.items())) + ')'
+	return 'list(' + ','.join(['%s=%s' % (Str4R(a[0]), Str4R(a[1])) for a in obj.items()]) + ')'
 
 def OtherStr(obj):
 	if has_numpy:
@@ -147,11 +202,11 @@ def OtherStr(obj):
 			def getVec(ary):
 				tp = ary.dtype.kind
 				rlt = ary.reshape(reduce(lambda a,b=1:a*b, ary.shape))
-				rlt = tp == 'b' and map(lambda a:a and 'TRUE' or 'FALSE', rlt) or rlt.tolist()
+				rlt = tp == 'b' and [a and 'TRUE' or 'FALSE' for a in rlt] or rlt.tolist()
 				if tp != 'V':
 					return tpdic.get(tp, 'c(%s)') % repr(rlt)[1:-1]
 				# record array
-				rlt = map(SeqStr, rlt) # each record will be mapped to vector or list
+				rlt = list(map(SeqStr, rlt)) # each record will be mapped to vector or list
 				return tpdic.get(tp, 'list(%s)') % (', '.join(rlt)) # use str here instead of repr since it has already been converted to str by SeqStr
 
 			if len(shp) == 1: # to vector
@@ -169,7 +224,7 @@ def OtherStr(obj):
 			else: # to array
 				dim = list(shp[-2:]) # row, col
 				dim.extend(shp[-3::-1])
-				newaxis = range(len(shp))
+				newaxis = list(range(len(shp)))
 				newaxis[-2:] = [len(shp)-1, len(shp)-2]
 				return 'array(%s, dim=c(%s))' % (getVec(obj.transpose(newaxis)), repr(dim)[1:-1])
 			# record array and char array
@@ -182,7 +237,7 @@ def OtherStr(obj):
 
 base_tps = [type(None), bool, int, long, float, complex, str, unicode, list, tuple, set, frozenset, dict] # use type(None) instead of NoneType since the latter cannot be found in the types module in Python 3
 base_tps.reverse()
-str_func = {type(None):NoneStr, bool:BoolStr, int:repr, long:LongStr, float:repr, complex:ComplexStr, str:repr, unicode:repr, list:SeqStr, tuple:SeqStr, set:SeqStr, frozenset:SeqStr, dict:DictStr}
+str_func = {type(None):NoneStr, bool:BoolStr, long:LongStr, int:repr, float:repr, complex:ComplexStr, str:repr, unicode:repr, list:SeqStr, tuple:SeqStr, set:SeqStr, frozenset:SeqStr, dict:DictStr}
 
 def Str4R(obj):  
 	'''
@@ -326,16 +381,10 @@ class R: # (object):
 	''' 
 	_DEBUG_MODE = True
 
-	def __init__(self, RCMD='R', wait0=2, wait=0.01, max_len=1000, use_numpy=True, use_dict=None, host='localhost', user=None, ssh='ssh', return_err=True):
+	def __init__(self, RCMD='R', max_len=1000, use_numpy=True, use_dict=None, host='localhost', user=None, ssh='ssh', return_err=True):
 		'''
 		RCMD: The name of a R interpreter, path information should be included
 			if it is not in the system search path.
-		wait0: The time span for recv_some to receive data from pipe for the
-			first time
-		wait: The time span for recv_some to receive data from pipe max_len:
-			define the upper limitation for the length of command string.  A
-			command string will be passed to R by a temporary file if it is
-			longer than this value.
 		use_numpy: Used as a boolean value. A False value will disable numpy
 			even if it has been imported.
 		use_dict: A R named list will be returned as a Python dictionary if
@@ -367,12 +416,10 @@ class R: # (object):
 		# underscore (DunderAlias) at both ends of the name) may look at the
 		# class rather than the instance even for old-style classes."
 		self.__dict__.update({
-			'wait0' : wait0, 
-			'wait' : wait, 
 			'max_len' : max_len, 
 			'use_dict' : use_dict,
 			'localhost' : host=='localhost', 
-			'tail' : sys.platform=='win32' and '\r\n' or '\n'})
+			'newline' : sys.platform=='win32' and '\r\n' or '\n'})
 
 		RCMD = [RCMD] #shlex.split(RCMD) - shlex do not work properly on Windows! #re.split(r'\s', RCMD)
 		if not self.localhost:
@@ -384,48 +431,46 @@ class R: # (object):
 		args = ('--quiet', '--no-save', '--no-restore') # "--slave" cannot be used on Windows!
 		for arg in args:
 			if arg not in RCMD: RCMD.append(arg)
-		if hasattr(subprocess, 'STARTUPINFO'):
+		if _has_subp and hasattr(subprocess, 'STARTUPINFO'):
 			info = subprocess.STARTUPINFO()
 			info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
 			info.wShowWindow = subprocess.SW_HIDE
 		else: info = None
 		self.__dict__.update({
-			'prog' : Popen(RCMD, stdin=PIPE, stdout=PIPE, stderr=return_err and subprocess.STDOUT or None, startupinfo=info), 
+			'prog' : Popen(RCMD, stdin=PIPE, stdout=PIPE, stderr=return_err and _STDOUT or None, startupinfo=info), 
 			'has_numpy' : use_numpy and has_numpy, 
 			'Rfun' : self.__class__.__Rfun % ((use_numpy and has_numpy) and 'TRUE' or 'FALSE')})
 		self.__call__(self.Rfun)
-		to_discard = recv_some(self.prog, e=0, t=wait0)
+		#to_discard = recv_some(self.prog, e=0, t=wait0)
 
 	def __runOnce(self, CMD, use_try=None):
 		'''
 		CMD: a R command string
 		'''
 		use_try = use_try or self._DEBUG_MODE
-		tail = self.tail
+		newline = self.newline
 		tail_token = 'R command at time: %s' % repr(time.time())
-		tail_cmd = 'print("%s")%s' % (tail_token, tail)
-		rlt_tail = '> print("%s")%s[1] "%s"%s> ' % (tail_token, tail, tail_token, tail)
-		rlt_tail1 = rlt_tail.replace('\r', '') # sometimes Windows use '\n' only, e.g. after "library(fastICA)"!
-		#if not CMD.endswith(tail): CMD = CMD + tail
+		#tail_token_r = re.sub(r'[\(\)\.]', r'\\\1', tail_token)
+		tail_cmd = 'print("%s")%s' % (tail_token, newline)
+		re_tail = re.compile(r'>\sprint\("%s"\)\r?\n\[1\]\s"%s"\r?\n$' % (tail_token.replace(' ', '\\s'), tail_token.replace(' ', '\\s')) )
 		if len(CMD) <= self.max_len or not self.localhost: 
 			fn = None
 		else:
 			fh, fn = tempfile.mkstemp()
-			os.fdopen(fh, 'wb').write(CMD)
+			os.fdopen(fh, 'wb').write(_mybytes(CMD))
+			if sys.platform == 'cli': os.close(fh) # this is necessary on IronPython 
 			CMD = 'source("%s")' % fn.replace('\\', '/')
-		CMD = (use_try and 'try({%s})%s%s' or '%s%s%s') % (CMD, tail, tail_cmd)
-		send_all(self.prog, CMD)
+		CMD = (use_try and 'try({%s})%s%s' or '%s%s%s') % (CMD, newline, tail_cmd)
+		sendAll(self.prog, CMD)
 		rlt = ''
-		while not rlt.endswith(rlt_tail) and not rlt.endswith(rlt_tail1):
+		while not re_tail.search(rlt): 
 			try:
-				rltonce = recv_some(self.prog, e=1, t=self.wait)
+				rltonce = readLine(self.prog)
 				if rltonce: rlt = rlt + rltonce
 			except: break
 		else: 
-			if rlt.endswith(rlt_tail):
-				rlt = rlt[:-len(rlt_tail)]
-			elif rlt.endswith(rlt_tail1):
-				rlt = rlt[:-len(rlt_tail1)]
+			rlt = re_tail.sub('', rlt)
+			if rlt.startswith('> '): rlt = rlt[2:]
 		if fn is not None:
 			os.unlink(fn)
 		return rlt
@@ -462,14 +507,12 @@ class R: # (object):
 		if use_dict is None: use_dict = self.use_dict
 		cmd = '.getRvalue4Python__(%s, use_dict=%s)' % (obj, use_dict is None and 'NULL' or use_dict and 'TRUE' or 'FALSE')
 		rlt = self.__call__(cmd, use_try=use_try)
-		head = (use_try and 'try({%s})%s[1] ' or '%s%s[1] ') % (cmd, self.tail) 
+		head = (use_try and 'try({%s})%s[1] ' or '%s%s[1] ') % (cmd, self.newline) 
 		# sometimes (e.g. after "library(fastICA)") the R on Windows uses '\n' instead of '\r\n'
 		head = rlt.startswith(head) and len(head) or len(head) - 1 
-		tail = rlt.endswith(self.tail) and len(rlt) - len(self.tail) or len(rlt) - len(self.tail) + 1 # - len('"')
+		tail = rlt.endswith(self.newline) and len(rlt) - len(self.newline) or len(rlt) - len(self.newline) + 1 # - len('"')
 		try:
 			rlt = eval(eval(rlt[head:tail])) # The inner eval remove quotes and recover escaped characters.
-			#head, tail = head + 1, tail - 1 # to remove quote
-			#rlt = eval(rlt[head:tail])
 		except:
 			raise RError(rlt)
 		return rlt
@@ -491,7 +534,7 @@ class R: # (object):
 		self.__call__('rm(%s)' % obj)
 
 	def __del__(self):
-		send_all(self.prog, 'q("no")'+self.tail)
+		sendAll(self.prog, 'q("no")'+self.newline)
 		self.prog = None
 
 	def __getattr__(self, obj, use_dict=None): # to model r.XXX
@@ -544,15 +587,12 @@ class R: # (object):
 
 
 # for a single-round duty:
-def runR(CMDS, Robj='R', wait0=2, wait=0.01, max_len=1000, use_numpy=True, use_dict=None, host='localhost', user=None, ssh='ssh'):
+def runR(CMDS, Robj='R', max_len=1000, use_numpy=True, use_dict=None, host='localhost', user=None, ssh='ssh'):
 	'''
 	Run a (list of) R command(s), and return the output from the STDOUT.
 
 	CMDS: a R command string or a list of R commands.
 	Robj: can be a shell command (like /usr/bin/R), or the R class.
-	wait0: The time span for recv_some to receive data from pipe for the first
-		time.
-	wait: The time span for recv_some to receive data from pipe.
 	max_len: define the upper limitation for the length of command string. A
 		command string will be passed to R by a temporary file if it is longer
 		than this value.
@@ -574,7 +614,7 @@ def runR(CMDS, Robj='R', wait0=2, wait=0.01, max_len=1000, use_numpy=True, use_d
 	ssh: The program to login to remote computer.
 	'''
 	if isinstance(Robj, basestring):
-		Robj = R(RCMD=Robj, wait0=wait0, wait=wait, max_len=max_len, use_numpy=use_numpy, use_dict=use_dict, host=host, user=user, ssh=ssh)
+		Robj = R(RCMD=Robj, max_len=max_len, use_numpy=use_numpy, use_dict=use_dict, host=host, user=user, ssh=ssh)
 	rlt = Robj.run(CMDS=CMDS)
 	if len(rlt) == 1: rlt = rlt[0]
 	return rlt
