@@ -57,6 +57,8 @@ if __name__ == '__main__':
 	a = range(10000) #00)
 	sa = 'a <- ' + Str4R(a)
 	rlt = runR(sa)
+	print(rlt)
+	print('\nTest passed!\n\n')
 
 	# to use an R on remote server, you need to provide correct parameter to initialize the R instance:
 	# rsrv = R(RCMD='/usr/local/bin/R', host='My_server_name_or_IP', user='username')
